@@ -19,19 +19,21 @@ pub fn is_path_in_directory(parent: &Path, path: &Path) -> Result<bool> {
     Ok(canonical_path.starts_with(canonical_parent))
 }
 
-/// Create a file with the content given
-pub fn create_file(path: &Path, content: &str) -> Result<()> {
-    let mut file =
-        File::create(path).with_context(|| format!("Failed to create file {}", path.display()))?;
-    file.write_all(content.as_bytes())?;
+/// Creates the parent of a directory, if needed.
+fn create_parent(path: &Path) -> Result<()> {
+    if let Some(parent) = path.parent() {
+        create_directory(parent)?;
+    }
     Ok(())
 }
 
-/// Create a directory at the given path if it doesn't exist already
-pub fn ensure_directory_exists(path: &Path) -> Result<()> {
-    if !path.exists() {
-        create_directory(path)?;
-    }
+/// Create a file with the content given
+/// `content`` can be `&str`, `String`, or `&String` (and probably others)
+pub fn create_file(path: &Path, content: impl AsRef<str>) -> Result<()> {
+    create_parent(path)?;
+    let mut file =
+        File::create(path).with_context(|| format!("Failed to create file {}", path.display()))?;
+    file.write_all(content.as_ref().as_bytes())?;
     Ok(())
 }
 
@@ -66,12 +68,7 @@ pub fn copy_file(src: &Path, dest: &Path, base_path: &Path, hard_link: bool) ->
     let relative_path = src.strip_prefix(base_path).unwrap();
     let target_path = dest.join(relative_path);
 
-    if let Some(parent_directory) = target_path.parent() {
-        create_dir_all(parent_directory).with_context(|| {
-            format!("Failed to create directory {}", parent_directory.display())
-        })?;
-    }
-
+    create_parent(&target_path)?;
     copy_file_if_needed(src, &target_path, hard_link)
 }
 
@@ -80,11 +77,7 @@ pub fn copy_file(src: &Path, dest: &Path, base_path: &Path, hard_link: bool) ->
 /// 2. Its modification timestamp is identical to that of the src file.
 /// 3. Its filesize is identical to that of the src file.
 pub fn copy_file_if_needed(src: &Path, dest: &Path, hard_link: bool) -> Result<()> {
-    if let Some(parent_directory) = dest.parent() {
-        create_dir_all(parent_directory).with_context(|| {
-            format!("Failed to create directory {}", parent_directory.display())
-        })?;
-    }
+    create_parent(&dest)?;
 
     if hard_link {
         if dest.exists() {
